// Utena IIIMP Server for Anthy
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iconv.h>
#include <list>
#include <anthy.h>
#include "rkconv.h"
#include "rkmap.h"
#include "imconn.h"
#include "packet.h"

#define PORT_NUMBER 9010

int trace_flag;
static pid_t pid;
static int background_p;

RKMap *rk_hira_map;
RKMap *rk_kata_map;

static int port_number=PORT_NUMBER;
static int serv_fd;
static list<IMConn *> ims;

extern void initKeyAtoms();

static void parse_args(int argc,char **argv);
static void print_usage();
static void print_version();
static int event_loop();
static int accept_proc();
static void kill_prev();

static int init_serv_fd();

#if __GLIBC_MINOR__ == 1
#define ICONV_ARG2 (const char **)
#else
#define ICONV_ARG2
#endif

STRING createSTRING(char *s)
{
  STRING str;
  iconv_t conv;

  conv = iconv_open("UTF-16","EUC-JP");
  if ( (int)conv == -1 ){
    printf("Failed to create iconv.\n");
    exit(0);
  }

  size_t ilen,olen,buflen,ret,len;
  int i;
  len = strlen(s);
  ilen = len;
  buflen = ilen*5;
  char *outbuf = (char *)alloca(buflen);
  char *pp = outbuf;
  ret = iconv(conv,ICONV_ARG2 &s,&ilen,&outbuf,&buflen);
  if ( ret == -1 ){
    printf("failed iconv.\n");
    return str;
  }
  unsigned short *p;
  p = (unsigned short*)pp;
  olen = len*5 - buflen;

  for ( i = 0 ; i < olen /2 ; i++){
    // UTF is BIG ENDIAN?
    str.push_back((p[i]>>8)|((p[i]<<8)&65280));
  }
  iconv_close(conv);
  return str;
}

void kill_prev()
{
  char *s;
  s = getenv("UTENA_PID");
  if ( s ){
    pid_t p = atoi(s);
    if ( p ){
      kill(p,SIGKILL);
    }
  }else{
    printf("Can't find Utena. (in this world)\n");
  }
  exit(0);
}

int init_serv_fd()
{
    int foo;
    int fd;
    struct sockaddr_in myhost;

    fd=socket(AF_INET,SOCK_STREAM,0);
    if ( fd==-1 ){
	perror("failed in socket()");
	return -1;
    }

    foo = 1;
    foo = setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,&foo ,4);
    if ( foo == -1){
        perror("failed to set reuse option");
    }

    myhost.sin_family=AF_INET;
    myhost.sin_port=htons(port_number);
    myhost.sin_addr.s_addr=htonl(INADDR_ANY);

    foo = bind(fd,(struct sockaddr *)&myhost,16);
    if(foo==-1){
	perror("failed in bind()");
	return -1;
    }
    foo = listen(fd,5);
    if ( foo == -1){
	perror("failed in listen()");
	return -1;
    } 
    return fd;
}

void print_version()
{
  printf("UTENA(Anthy-server)-"VERSION"\n");
  exit(0);
}

void print_usage()
{
  printf("UTENA(Anthy-server)-"VERSION".\n"
	 " --trace: trace connection.\n"
	 " --version: show version.\n"
	 " --help: show this help.\n"
	 " --background: run background.\n"
	 " --kill: kill previous process.\n");
  exit(0);
}

void parse_args(int argc,char **argv)
{
  int i;
  for ( i = 1 ; i < argc ; i++){
    if ( !strncmp(argv[i],"--",2)){
      char *arg = &argv[i][2];
      if ( !strcmp(arg,"trace")){
	trace_flag = 1;
      }else if (!strcmp(arg,"help")){
	print_usage();
      }else if (!strcmp(arg,"version")){
	print_version();
      }else if (!strcmp(arg,"background")){
	background_p = 1;
      }else if (!strcmp(arg,"kill")){
	kill_prev();
      }else{
	print_usage();
      }
    }
  }
}

int accept_proc()
{
  socklen_t len;
  struct sockaddr_in peer;
  len = sizeof(sockaddr_in);
  int new_fd = accept(serv_fd,(struct sockaddr *)&peer,&len);
  if ( new_fd == -1 ){
    perror("Failed to accept.\n");
    return -1;
  }
  if ( trace_flag ){
    printf("Accept\n");
  }
  IMConn *conn = new IMConn(new_fd);
  ims.push_back(conn);
  return 0;
}

int event_loop()
{
  /* setup poll table */
  struct pollfd *polltab;
  int nr_fd = 1,nth;
  nr_fd += ims.size();
  polltab = (struct pollfd *)alloca(sizeof(struct pollfd)*nr_fd);
  /* setup polltab for fd to accept */
  nth = 0;
  polltab[nth].fd = serv_fd;
  polltab[nth].events = POLLIN;
  nth++;
  /* setup polltab for each connection */
  list<IMConn *>::iterator it;
  for ( it = ims.begin() ; it != ims.end() ; it ++){
    polltab[nth].fd = (*it)->getFd();
    polltab[nth].events = POLLIN;
    if ( (*it)->isWritePending()){
      polltab[nth].events |= POLLOUT;
    }
    nth++;
  }
  /* do poll */
  poll(polltab,nth,1000);
  /* check result for each connection */
  nth = 1;
  for ( it = ims.begin() ; it != ims.end() ; it ++){
    if ( polltab[nth].revents & POLLIN ){
      (*it)->read_proc();
    }
    if ( polltab[nth].revents & POLLOUT ){
      (*it)->write_proc();
    }
    nth ++;
  }
  /* reap invalid connection */
  for ( it = ims.begin() ; it != ims.end() ; ){
    if ( (*it)->isValid() ){
      it++;
    }else{
      if ( trace_flag ){
        printf("disconnect\n");
      }
      IMConn *im = *it;
      it = ims.erase(it);
      delete im;
    }
  }
  /* check result for fd to accept
     this shuld be after the check of each connection*/
  if ( polltab[0].revents & POLLIN ){
    /* Do accept */
    return accept_proc();
  }
  return 0;
}

int main(int argc,char **argv)
{
  int newfd,addr_len,pid;
  struct sockaddr_in peer;

  parse_args(argc,argv);

  if ( !background_p ){
    pid = getpid();
  }else{
    pid = fork();
    if ( pid == -1 ){
      perror("Failed to do fork()");
      return -1;
    }
  }

  if ( pid > 0 && background_p ){
    printf("UTENA_PID=%d ;export UTENA_PID\n",pid);
    printf("echo Utena pid=%d\n",pid);
    fflush(stdout);
    return -1;
  }

  initKeyAtoms();

  if ( anthy_init() == -1 ){
    printf("failed to init anthy.\n");
    return -1;
  }
  rk_hira_map = new RKMap(hiragana);
  rk_kata_map = new RKMap(katakana);
  serv_fd = init_serv_fd();
  if ( serv_fd == -1){
    exit(-1);
  }
  if ( trace_flag ){
    printf("Begin\n");
  }
  while(!event_loop());
  return 0;
}
